; #INDEX# =======================================================================================================================
; Title .........: AutoIt error logger (AUERLO) v2.0.0
; AutoIt Version : 3.3.14.3+
; Description ...: File- and stdout/-err (console) output of messages, regular- and COM Object errors, exit method and -code, etc.
; Author(s) .....: user4157124
; Link ..........: https://www.autoitscript.com/forum/topic/198929-autoit-error-logger-auerlo-udf/
; License .......: GNU Affero General Public License (GNU AGPL)
; ===============================================================================================================================
; This program is free software: you can redistribute it and/or modify
; it under the terms of the GNU Affero General Public License as
; published by the Free Software Foundation, either version 3 of the
; License, or (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU Affero General Public License for more details.
;
; You should have received a copy of the GNU Affero General Public License
; along with this program.  If not, see <https://www.gnu.org/licenses/>.
; ===============================================================================================================================
#include-once
#include <AutoItConstants.au3>; $UBOUND_..., $SOUND_NOWAIT
#include <FileConstants.au3>;   $FV_FILEVERSION
#include <WinAPI.au3>;          _WinAPI_GetLastError(), _WinAPI_GetLastErrorMessage()
#include <Date.au3>;            __AUERLO_GetTimeUTC() x5
#include <File.au3>;            _FileReadToArray()
#include <Array.au3>;           _ArrayDisplay(), _ArrayToString()
#include <EventLog.au3>;        __AUERLO_Output_WinEventLog() x3
#include "auerloConstants.au3"
#include "auerloGlobalTxt.au3"
#include "auerloGlobalOpt.au3"

#Region UDF

; #INDEX# =======================================================================================================================
; _AUERLO_Set
; _AUERLO_Log
; _AUERLO_LogAlt
; _AUERLO_LogWinAPI
; _AUERLO_LogEnv
; _AUERLO_Assert
; _AUERLO_FileClear
; _AUERLO_FileRead
; _AUERLO_View
; ===============================================================================================================================

; #FUNCTION# ====================================================================================================================
; Name ..........: _AUERLO_Set
; Description ...: Set option.
; Syntax ........: _AUERLO_Set($iOption[, $vData = Default])
; Parameters ....: $iOption             - $AUERLO_OPT__ENUM.
;                  $vData               - [optional] a variant value. Default is Default.
; Return values .: $AUERLO_RET__ENUM, @error $AUERLO_ERR__ENUM.
; Remarks .......:
; ===============================================================================================================================
Func _AUERLO_Set(Const $iOption, $vData = Default)

	__AUERLO_CheckOpt($iOption)
	If @error Then Return SetError(@error, @extended, $AUERLO_RET_ERR)

	If $vData = Default Then $vData = $__g_aAUERLO_Opt[$AUERLO_SET_DEF][$iOption]

	Switch $iOption

		Case $AUERLO_OPT_FILEPATH
			If Not StringLen($vData) Then Return SetError($AUERLO_ERR_DATA, 0, $AUERLO_RET_ERR)
			$vData = StringReplace($vData, '/', '\')

		Case $AUERLO_OPT_FILEDELMCOL To $AUERLO_OPT_FILEDELMROW
			If Not StringLen($vData) Then Return SetError($AUERLO_ERR_DATA, 0, $AUERLO_RET_ERR)

		Case $AUERLO_OPT_FILEENC
			If $vData < $FO_UTF16_LE Or $vData > $FO_FULLFILE_DETECT Then Return SetError($AUERLO_ERR_DATA, 0, $AUERLO_RET_ERR)

		Case $AUERLO_OPT_LVLFILE To $AUERLO_OPT_LVLCUSTOM; $AUERLO_OPT_LVLFILE, $AUERLO_OPT_LVLSTD, $AUERLO_OPT_LVLNOTIFY, $AUERLO_OPT_LVLPROMPT, $AUERLO_OPT_LVLWEL, $AUERLO_OPT_LVLCUSTOM
			__AUERLO_CheckLevel($vData)
			If @error Then Return SetError(@error, @extended, $AUERLO_RET_ERR)

		Case $AUERLO_OPT_ACTIVE To $AUERLO_OPT_LOGCUSTOM; $AUERLO_OPT_ACTIVE, $AUERLO_OPT_TIMEUTC, $AUERLO_OPT_STDCOLOR, $AUERLO_OPT_LOGFILE, $AUERLO_OPT_LOGSTD, $AUERLO_OPT_LOGNOTIFY, $AUERLO_OPT_LOGPROMPT, $AUERLO_OPT_LOGWEL, $AUERLO_OPT_LOGCUSTOM
			$vData = ($vData ? True : False); Decoration; whatever contained evaluates to either True or False already.

		Case $AUERLO_OPT_FNCTIME
			If Not IsFunc($vData)      Then Return SetError($AUERLO_ERR_FNCDEF, 0, $AUERLO_RET_ERR)
			If Not $vData() Or @error  Then Return SetError($AUERLO_ERR_FNCRET, 0, $AUERLO_RET_ERR)

		Case $AUERLO_OPT_FNCCOM
			If Not ($vData = Null) Then

				If Not IsFunc($vData) Then Return SetError($AUERLO_ERR_FNCDEF, 0, $AUERLO_RET_ERR)

				$__g_aAUERLO_Opt[$AUERLO_SET_CUR][$iOption] = ObjEvent('AutoIt.Error', FuncName($vData))
				If @error Then _
				Return SetError($AUERLO_ERR_FNCREG, @error, $AUERLO_RET_ERR)
				Return SetError($AUERLO_ERR_OK,     0,      $AUERLO_RET_OK)

			EndIf

		Case $AUERLO_OPT_FNCEXIT
			If $vData = Null Then

				If $__g_aAUERLO_Opt[$AUERLO_SET_CUR][$iOption] Then

					If Not OnAutoItExitUnRegister($__g_aAUERLO_Opt[$AUERLO_SET_CUR][$iOption]) Then _
					Return SetError($AUERLO_ERR_FNCUNR, 0, $AUERLO_RET_ERR)

				EndIf

			ElseIf Not ($__g_aAUERLO_Opt[$AUERLO_SET_CUR][$iOption] = $vData) Then

				If Not IsFunc($vData) Then Return SetError($AUERLO_ERR_FNCDEF, 0, $AUERLO_RET_ERR)

				If Not ($__g_aAUERLO_Opt[$AUERLO_SET_CUR][$iOption] = Null) _
				And Not OnAutoItExitUnRegister($__g_aAUERLO_Opt[$AUERLO_SET_CUR][$iOption]) Then _
				Return SetError($AUERLO_ERR_FNCUNR, 0, $AUERLO_RET_ERR)

				If Not OnAutoItExitRegister($vData) Then _
				Return SetError($AUERLO_ERR_FNCREG, 0, $AUERLO_RET_ERR)

			EndIf

		Case $AUERLO_OPT_FNCFILE To $AUERLO_OPT_FNCWEL; $AUERLO_OPT_FNCFILE, $AUERLO_OPT_FNCSTD, $AUERLO_OPT_FNCNOTIFY, $AUERLO_OPT_FNCPROMPT, $AUERLO_OPT_FNCWEL
			If Not IsFunc($vData) Then Return SetError($AUERLO_ERR_FNCDEF, 0, $AUERLO_RET_ERR)

		Case $AUERLO_OPT_FNCCUSTOM
			If Not ($vData = Null) And Not IsFunc($vData) Then Return SetError($AUERLO_ERR_FNCDEF, 0, $AUERLO_RET_ERR)

	EndSwitch

	$__g_aAUERLO_Opt[$AUERLO_SET_CUR][$iOption] = $vData

	Return SetError($AUERLO_ERR_OK, 0, $AUERLO_RET_OK)
EndFunc

; #FUNCTION# ====================================================================================================================
; Name ..........: _AUERLO_Log
; Description ...: Create log item.
; Syntax ........: _AUERLO_Log([$sMessage = ''[, $iLevel = Default[, $bReset = Default[, $iError = @error[, $iExtended = @extended[, $iLine = @ScriptLineNumber[, $sFilePath = @ScriptFullPath]]]]]]])
; Parameters ....: $sMessage            - [optional] a string value. Default is ''.
;                  $iLevel              - [optional] $AUERLO_LVL__ENUM. Default is Default.
;                  $bReset              - [optional] a boolean value. Default is Default. Returns 0 for @error and @extended if True.
;                  $iError              - [optional] an integer value. Default is @error.
;                  $iExtended           - [optional] an integer value. Default is @extended.
;                  $iLine               - [optional] an integer value. Default is @ScriptLineNumber.
;                  $sFilePath           - [optional] a string value. Default is @ScriptFullPath.
; Return values .: $AUERLO_RET__ENUM, @error $iError, @extended $iExtended.
; Remarks .......: Passes @error and @extended.
; ===============================================================================================================================
Func _AUERLO_Log(Const $sMessage = '', $iLevel = Default, Const $bReset = Default, Const $iError = @error, Const $iExtended = @extended, Const $iLine = @ScriptLineNumber, Const $sFilePath = @ScriptFullPath)
	Local $sDateTime = $__g_aAUERLO_Opt[$AUERLO_SET_CUR][$AUERLO_OPT_FNCTIME]()

	If Not $__g_aAUERLO_Opt[$AUERLO_SET_CUR][$AUERLO_OPT_ACTIVE] Then

		If $bReset Then _
		Return SetError($AUERLO_ERR_OK, 0,          $AUERLO_RET_OK)
		Return SetError($iError,        $iExtended, $AUERLO_RET_OK)

	EndIf

	If $iLevel = Default                          Then $iLevel = $AUERLO_LVL_TRACE
	If $iExtended And $iLevel < $AUERLO_LVL_DEBUG Then $iLevel = $AUERLO_LVL_DEBUG
	If $iError    And $iLevel < $AUERLO_LVL_ERROR Then $iLevel = $AUERLO_LVL_ERROR

	__AUERLO_CheckLevel($iLevel)
	If @error Then $iLevel = $AUERLO_LVL_ERROR

	Local $aData[$AUERLO_ATR__ENUM]
		  $aData[$AUERLO_ATR_DATETIME] = __AUERLO_GetTxtStripped($sDateTime)
		  $aData[$AUERLO_ATR_MESSAGE]  = __AUERLO_GetTxtStripped($sMessage)
		  $aData[$AUERLO_ATR_FILENAME] = __AUERLO_GetTxtStripped($sFilePath)
		  $aData[$AUERLO_ATR_LEVEL]    = Int($iLevel)
		  $aData[$AUERLO_ATR_LINE]     = Int($iLine)
		  $aData[$AUERLO_ATR_ERROR]    = Int($iError)
		  $aData[$AUERLO_ATR_EXTENDED] = Int($iExtended)

	__AUERLO_Dispatch($aData)

	If $bReset Then _
	Return SetError($AUERLO_ERR_OK, 0,          $AUERLO_RET_OK)
	Return SetError($iError,        $iExtended, $AUERLO_RET_OK)
EndFunc

; #FUNCTION# ====================================================================================================================
; Name ..........: _AUERLO_LogAlt
; Description ...: Create log item. Function calls in $sMessage/$iLevel/$bReset parameters without affecting @error and @extended.
; Syntax ........: _AUERLO_LogAlt([$iError = @error[, $iExtended = @extended[, $sMessage = ''[, $iLevel = Default[, $bReset = Default[, $iLine = @ScriptLineNumber[, $sFilePath = @ScriptFullPath]]]]]]])
; Parameters ....: $iError              - [optional] an integer value. Default is @error.
;                  $iExtended           - [optional] an integer value. Default is @extended.
;                  $sMessage            - [optional] a string value. Default is ''.
;                  $iLevel              - [optional] $AUERLO_LVL__ENUM. Default is Default.
;                  $bReset              - [optional] a boolean value. Default is Default. Returns 0 for @error and @extended if True.
;                  $iLine               - [optional] an integer value. Default is @ScriptLineNumber.
;                  $sFilePath           - [optional] a string value. Default is @ScriptFullPath.
; Return values .: $AUERLO_RET__ENUM, @error $iError, @extended $iExtended.
; Remarks .......: Passes @error and @extended. _AUERLO_Log() wrapper.
; ===============================================================================================================================
Func _AUERLO_LogAlt(Const $iError = @error, Const $iExtended = @extended, Const $sMessage = '', Const $iLevel = Default, Const $bReset = Default, Const $iLine = @ScriptLineNumber, Const $sFilePath = @ScriptFullPath)
	Local Const $iRet = _AUERLO_Log($sMessage, $iLevel, $bReset, $iError, $iExtended, $iLine, $sFilePath)
	Return SetError(@error, @extended, $iRet)
EndFunc

; #FUNCTION# ====================================================================================================================
; Name ..........: _AUERLO_LogWinAPI
; Description ...: Create log item. Records _WinAPI_GetLastError() and _WinAPI_GetLastErrorMessage().
; Syntax ........: _AUERLO_LogWinAPI([$bReset = Default[, $iError = @error[, $iExtended = @extended[, $iLine = @ScriptLineNumber[, $sFilePath = @ScriptFullPath]]]]])
; Parameters ....: $bReset              - [optional] a boolean value. Default is Default. Returns 0 for @error and @extended if True.
;                  $iError              - [optional] an integer value. Default is @error.
;                  $iExtended           - [optional] an integer value. Default is @extended.
;                  $iLine               - [optional] an integer value. Default is @ScriptLineNumber.
;                  $sFilePath           - [optional] a string value. Default is @ScriptFullPath.
; Return values .: $AUERLO_RET__ENUM, @error $iError, @extended $iExtended.
; Remarks .......: Passes @error and @extended.
; ===============================================================================================================================
Func _AUERLO_LogWinAPI(Const $bReset = Default, Const $iError = @error, Const $iExtended = @extended, Const $iLine = @ScriptLineNumber, Const $sFilePath = @ScriptFullPath)
	Local Const $iRet = _AUERLO_Log('__AUERLO_LogWinAPI() - ' & _WinAPI_GetLastErrorMessage(), Default, True, _WinAPI_GetLastError(), 0, $iLine, $sFilePath)

	If $bReset Then _
	Return SetError($AUERLO_ERR_OK, 0,          $iRet)
	Return SetError($iError,        $iExtended, $iRet)
EndFunc

; #FUNCTION# ====================================================================================================================
; Name ..........: _AUERLO_LogEnv
; Description ...: Create log item. Records environment information.
; Syntax ........: _AUERLO_LogEnv([$iLevel = Default[, $iLine = @ScriptLineNumber[, $sFilePath = @ScriptFullPath]]])
; Parameters ....: $iLevel              - [optional] $AUERLO_LVL__ENUM. Default is Default.
;                  $iLine               - [optional] an integer value. Default is @ScriptLineNumber.
;                  $sFilePath           - [optional] a string value. Default is @ScriptFullPath.
; Return values .: $AUERLO_RET__ENUM, @error $AUERLO_ERR__ENUM.
; Remarks .......:
; ===============================================================================================================================
Func _AUERLO_LogEnv($iLevel = Default, Const $iLine = @ScriptLineNumber, Const $sFilePath = @ScriptFullPath)
	Local Const $aItem = [ _
							@ComputerName, _
							@CPUArch, _
							@OSType, _
							@OSArch, _
							@OSVersion, _
							@OSBuild, _
							@OSServicePack, _
							@OSLang, _
							'AutoIt v' & @AutoItVersion, _
							@AutoItX64 ? 'X64'      : 'X86', _
							'File v'   & (@Compiled ? FileGetVersion(@ScriptFullPath, $FV_FILEVERSION) : '0.0.0.0'), _
							@Compiled  ? 'Compiled' : 'notCompiled', _
							IsAdmin()  ? 'Admin'    : 'notAdmin', _
							@ScriptFullPath, _
							@WorkingDir _
						 ]

	If $iLevel = Default Then $iLevel = $AUERLO_LVL_DEBUG
	Local Const $iRet = _AUERLO_Log('_AUERLO_LogEnv() - ' & _ArrayToString($aItem, '|'), $iLevel, True, $AUERLO_ERR_OK, 0, $iLine, $sFilePath)

	Return SetError(@error, @extended, $iRet)
EndFunc

; #FUNCTION# ====================================================================================================================
; Name ..........: _AUERLO_Assert
; Description ...: Create log item. Records assertion result.
; Syntax ........: _AUERLO_Assert($vAssertion[, $sMessage = ''[, $iLine = @ScriptLineNumber[, $sFilePath = @ScriptFullPath]]])
; Parameters ....: $vAssertion          - An expression.
;                  $sMessage            - [optional] a string value. Default is ''.
;                  $iLine               - [optional] an integer value. Default is @ScriptLineNumber.
;                  $sFilePath           - [optional] a string value. Default is @ScriptFullPath.
; Return values .: A boolean value, @error $AUERLO_ERR__ENUM.
; Remarks .......:
; Example .......: If Not _AUERLO_Assert(@DesktopWidth >= 640) Then Exit 0x7FFFF0B6
;                  _AUERLO_Assert(FileClose($hFile))
; ===============================================================================================================================
Func _AUERLO_Assert(Const $vAssertion, Const $sMessage = '', Const $iLine = @ScriptLineNumber, Const $sFilePath = @ScriptFullPath)
	Local Const $bResult = ($vAssertion ? True : False)

	_AUERLO_Log( _
		'_AUERLO_Assert(' & $bResult & ')' & ($sMessage ? ' - ' & $sMessage : ''), _
		$bResult ? $AUERLO_LVL_DEBUG : $AUERLO_LVL_WARNING, _
		True, _
		$AUERLO_ERR_OK, _
		0, _
		$iLine, _
		$sFilePath _
	)

	Return SetError(@error, @extended, $bResult)
EndFunc

; #FUNCTION# ====================================================================================================================
; Name ..........: _AUERLO_FileClear
; Description ...: Delete current error log file ($AUERLO_OPT_FILEPATH).
; Syntax ........: _AUERLO_FileClear()
; Parameters ....: None
; Return values .: $AUERLO_RET__ENUM, @error $AUERLO_ERR__ENUM.
; Remarks .......:
; ===============================================================================================================================
Func _AUERLO_FileClear()

	If Not FileExists($__g_aAUERLO_Opt[$AUERLO_SET_CUR][$AUERLO_OPT_FILEPATH]) Then _
	Return SetError($AUERLO_ERR_FILEEXIST, 0, $AUERLO_RET_ERR)

	If Not FileDelete($__g_aAUERLO_Opt[$AUERLO_SET_CUR][$AUERLO_OPT_FILEPATH]) Then _
	Return SetError($AUERLO_ERR_FILEDELETE, 0, $AUERLO_RET_ERR)

	Return SetError($AUERLO_ERR_OK, 0, $AUERLO_RET_OK)
EndFunc

; #FUNCTION# ====================================================================================================================
; Name ..........: _AUERLO_FileRead
; Description ...: Error log file to 2D array.
; Syntax ........: _AUERLO_FileRead(Byref $aResult[, $sFilePath = Default[, $bHeader = Default[, $sDelmCol = Default[, $iFileEnc = Default]]]])
; Parameters ....: $aResult             - [in/out] a 2D array ($AUERLO_ATR__ENUM columns) on success.
;                  $sFilePath           - [optional] a string value. Default is Default.
;                  $bHeader             - [optional] a boolean value. Default is Default. Adds column names as first row.
;                  $sDelmCol            - [optional] a string value. Default is Default.
;                  $iFileEnc            - [optional] an integer value. Default is Default. Mode file encoding flag as per FileOpen().
; Return values .: $AUERLO_RET__ENUM, @error $AUERLO_ERR__ENUM.
; Remarks .......:
; ===============================================================================================================================
Func _AUERLO_FileRead(ByRef $aResult, $sFilePath = Default, Const $bHeader = Default, Const $sDelmCol = Default, Const $iFileEnc = Default)

	If $sFilePath = Default Then $sFilePath = $__g_aAUERLO_Opt[$AUERLO_SET_CUR][$AUERLO_OPT_FILEPATH]
	If Not FileExists($sFilePath) Then Return SetError($AUERLO_ERR_FILEEXIST, 0, $AUERLO_RET_ERR)

	Local $iErr  = 0
	Local $hFile = FileOpen($sFilePath, BitOR($FO_READ, ($iFileEnc = Default ? $__g_aAUERLO_Opt[$AUERLO_SET_CUR][$AUERLO_OPT_FILEENC] : $iFileEnc)))
	If $hFile = -1 Then Return SetError($AUERLO_ERR_FILEOPEN, 0, $AUERLO_RET_ERR)

	_FileReadToArray( _
		$hFile, _
		$aResult, _
		$FRTA_NOCOUNT, _
		($sDelmCol = Default ? $__g_aAUERLO_Opt[$AUERLO_SET_CUR][$AUERLO_OPT_FILEDELMCOL] : $sDelmCol) _
	)
	$iErr = @error
	FileClose($hFile)

	If $iErr Then Return SetError($AUERLO_ERR_FILEREAD, $iErr, $AUERLO_RET_ERR)

	If $bHeader Then

		Local $aHeader[1][$AUERLO_ATR__ENUM]

		For $i1 = 0 To $AUERLO_ATR__ENUM - 1

			$aHeader[0][$i1] = $__g_aAUERLO_AtrTxt[$i1]

		Next

		_ArrayConcatenate($aHeader, $aResult, 0)
		If @error Then Return SetError($AUERLO_ERR_ARRAY, @error, $AUERLO_RET_ERR)

		$aResult = $aHeader

	EndIf

	Return SetError($AUERLO_ERR_OK, 0, $AUERLO_RET_OK)
EndFunc

; #FUNCTION# ====================================================================================================================
; Name ..........: _AUERLO_View
; Description ...: Display 2D array.
; Syntax ........: _AUERLO_View($aArray[, $iLevel = Default[, $sArrayRange = Default]])
; Parameters ....: $aArray              - 2D array (as produced by _AUERLO_FileRead() ).
;                  $iLevel              - [optional] $AUERLO_LVL__ENUM. Default is Default. Level filter (individual; not a threshold).
;                  $sArrayRange         - [optional] a string value. Default is Default. Range filter as per _ArrayDisplay().
; Return values .: $AUERLO_RET__ENUM, @error $AUERLO_ERR__ENUM.
; Remarks .......:
; ===============================================================================================================================
Func _AUERLO_View(Const $aArray, Const $iLevel = Default, $sArrayRange = Default)

	__AUERLO_CheckArray($aArray, 2, $AUERLO_ATR__ENUM)
	If @error Then Return SetError(@error, @extended, $AUERLO_RET_ERR)

	Local $iCount = 0
	Local $aResult[0][$AUERLO_ATR__ENUM]

	If $iLevel = Default Then

		$aResult = $aArray

	Else

		__AUERLO_CheckAttr($iLevel)
		If @error Then Return SetError(@error, @extended, $AUERLO_RET_ERR)

		For $i1 = 0 To UBound($aArray, $UBOUND_ROWS) - 1

			If Not ($aArray[$i1][$AUERLO_ATR_LEVEL] = $iLevel) Then ContinueLoop

			$iCount += 1
			ReDim $aResult[$iCount][$AUERLO_ATR__ENUM]

			For $i2 = 0 To $AUERLO_ATR__ENUM - 1

				$aResult[$iCount - 1][$i2] = $aArray[$i1][$i2]

			Next

		Next

	EndIf

	If $sArrayRange = Default Then $sArrayRange = ''

	For $i1 = 0 To UBound($aResult, $UBOUND_ROWS) - 1

		$aResult[$i1][$AUERLO_ATR_LEVEL]    = __AUERLO_GetTxtLevel($aResult[$i1][$AUERLO_ATR_LEVEL])
;		$aResult[$i1][$AUERLO_ATR_ERROR]    = Hex($aResult[$i1][$AUERLO_ATR_ERROR])
;		$aResult[$i1][$AUERLO_ATR_EXTENDED] = Hex($aResult[$i1][$AUERLO_ATR_EXTENDED])

	Next

	_ArrayDisplay($aResult, '_AUERLO_View()', $sArrayRange, BitOR($ARRAYDISPLAY_COLALIGNLEFT, $ARRAYDISPLAY_NOROW), Default, _ArrayToString($__g_aAUERLO_AtrTxt, '|'))
	If @error Then Return SetError($AUERLO_ERR_ARRAY, @error, $AUERLO_RET_ERR)

	Return SetError($AUERLO_ERR_OK, 0, $AUERLO_RET_OK)
EndFunc

#EndRegion

#Region Internal

Func __AUERLO_CheckArray(Const $aArray, Const $iDim, Const $iColumns = 0)

	If Not (UBound($aArray, $UBOUND_DIMENSIONS) = $iDim) _
	Or Not (UBound($aArray, $UBOUND_COLUMNS) = $iColumns) _
	Then Return SetError($AUERLO_ERR_ARRAY, 0, $AUERLO_RET_ERR)

	Return SetError($AUERLO_ERR_OK, 0, $AUERLO_RET_OK)
EndFunc

Func __AUERLO_CheckEnum(Const $iEnum, Const $iItem)

	Switch $iItem

		Case 0 To $iEnum - 1
			Return SetError($AUERLO_ERR_OK, 0, $AUERLO_RET_OK)

		Case Else
			Return SetError($AUERLO_ERR_ENUM, 0, $AUERLO_RET_ERR)

	EndSwitch

EndFunc

Func __AUERLO_CheckOpt(Const $iOption)

	__AUERLO_CheckEnum($AUERLO_OPT__ENUM, $iOption)
	If @error Then Return SetError($AUERLO_ERR_OPT, @extended, $AUERLO_RET_ERR)

	Return SetError($AUERLO_ERR_OK, 0, $AUERLO_RET_OK)
EndFunc

Func __AUERLO_CheckLevel(Const $iErrorLevel)

	__AUERLO_CheckEnum($AUERLO_LVL__ENUM, $iErrorLevel)
	If @error Then Return SetError($AUERLO_ERR_LVL, @extended, $AUERLO_RET_ERR)

	Return SetError($AUERLO_ERR_OK, 0, $AUERLO_RET_OK)
EndFunc

Func __AUERLO_CheckAttr(Const $iAttr)

	__AUERLO_CheckEnum($AUERLO_ATR__ENUM, $iAttr)
	If @error Then Return SetError($AUERLO_ERR_ATR, @extended, $AUERLO_RET_ERR)

	Return SetError($AUERLO_ERR_OK, 0, $AUERLO_RET_OK)
EndFunc

Func __AUERLO_GetTxtStripped($sStr)
	Local Const $sPlaceholder = ' '; Empty str (removal) may generate new (multi-char) delimiter (by removing character(s) in between).

	For $i1 = $AUERLO_OPT_FILEDELMCOL To $AUERLO_OPT_FILEDELMROW

		$sStr = StringReplace($sStr, $__g_aAUERLO_Opt[$AUERLO_SET_CUR][$i1], $sPlaceholder)

	Next

	Return SetError($AUERLO_ERR_OK, 0, $sStr)
EndFunc

Func __AUERLO_GetTxtLevel(Const $iLevel)

	__AUERLO_CheckLevel($iLevel)
	If @error Then Return SetError(@error, @extended, '')

	Return SetError($AUERLO_ERR_OK, 0, $__g_aAUERLO_LvlTxt[$iLevel])
EndFunc

Func __AUERLO_GetTxtExitcode(Const $iExitcode)

	If Not $iExitCode Then Return SetError($AUERLO_ERR_OK, 0, $iExitcode)

	Local Const $iResult = $iExitcode - 0x7FFFF068; Relation offset @exitCode to $__g_aAUERLO_ExcTxt -item.

	Switch $iResult

		Case 0 To UBound($__g_aAUERLO_ExcTxt, $UBOUND_ROWS) - 1
			Return SetError($AUERLO_ERR_OK, 0, $__g_aAUERLO_ExcTxt[$iResult])

		Case Else
			Return SetError($AUERLO_ERR_EXC, 0, $iExitcode)

	EndSwitch

EndFunc

Func __AUERLO_GetFormatLine($iColumns = Default, $sDelmCol = Default, $sDelmRow = Default)
	Local $sResult = ''

	If $iColumns = Default Then $iColumns = $AUERLO_ATR__ENUM
	If $sDelmCol = Default Then $sDelmCol = $__g_aAUERLO_Opt[$AUERLO_SET_CUR][$AUERLO_OPT_FILEDELMCOL]
	If $sDelmRow = Default Then $sDelmRow = $__g_aAUERLO_Opt[$AUERLO_SET_CUR][$AUERLO_OPT_FILEDELMROW]

	For $i1 = 0 To $iColumns - 1

		$sResult &= '%s' & ($i1 = $iColumns - 1 ? $sDelmRow : $sDelmCol)

	Next

	Return SetError($AUERLO_ERR_OK, 0, $sResult)
EndFunc

Func __AUERLO_GetTimeUTC($sDateTime = _NowCalc())
	Local $aDate, _
		  $aTime

	_DateTimeSplit($sDateTime, $aDate, $aTime)
	If @error Then Return SetError($AUERLO_ERR_DATETIME, @error, $sDateTime)

	Local $tDateTimeLoc = _Date_Time_EncodeSystemTime($aDate[2], $aDate[3], $aDate[1], $aTime[1], $aTime[2], $aTime[3])
	Local $pDateTimeLoc = DllStructGetPtr($tDateTimeLoc)
	If @error Then Return SetError($AUERLO_ERR_CONVERT, @error, $sDateTime)

	Local $tDateTimeUTC = _Date_Time_TzSpecificLocalTimeToSystemTime($pDateTimeLoc, 0)

	$sDateTime = _Date_Time_SystemTimeToDateTimeStr($tDateTimeUTC, 1)

	Return SetError($AUERLO_ERR_OK, 0, $sDateTime)
EndFunc

Func __AUERLO_GetTimeCurr()
	Local $sDateTime = _NowCalc()

	If Not $__g_aAUERLO_Opt[$AUERLO_SET_CUR][$AUERLO_OPT_TIMEUTC] Then _
	Return SetError($AUERLO_ERR_OK, 0, $sDateTime)

	$sDateTime = __AUERLO_GetTimeUTC($sDateTime)
	Return SetError(@error, @extended, $sDateTime)
EndFunc

Func __AUERLO_Dispatch(Const $aData); Const declaration denies registered functions ByRef access.
	Local Const $iOffsetFncToLvl = $AUERLO_OPT_LVLFILE - $AUERLO_OPT_FNCFILE, _; Relation offset $AUERLO_OPT_FNC... to $AUERLO_OPT_LVL... .
				$iOffsetFncToAct = $AUERLO_OPT_LOGFILE - $AUERLO_OPT_FNCFILE;    Relation offset $AUERLO_OPT_FNC... to $AUERLO_OPT_LOG... .

	For $i1 = $AUERLO_OPT_FNCFILE To $AUERLO_OPT_FNCCUSTOM

		If $__g_aAUERLO_Opt[$AUERLO_SET_CUR][$i1 + $iOffsetFncToAct] _
		And Not ($__g_aAUERLO_Opt[$AUERLO_SET_CUR][$i1] = Null) _
		And $aData[$AUERLO_ATR_LEVEL] >= $__g_aAUERLO_Opt[$AUERLO_SET_CUR][$i1 + $iOffsetFncToLvl] _
		Then $__g_aAUERLO_Opt[$AUERLO_SET_CUR][$i1]($aData)

	Next

	Return SetError($AUERLO_ERR_OK, 0, $AUERLO_RET_OK)
EndFunc

#EndRegion

#Region Input

Func __AUERLO_HandleExit()
	_AUERLO_Log( _
		'__AUERLO_HandleExit() - @exitMethod = "' & $__g_aAUERLO_MtdTxt[@exitMethod] & '". @exitCode = "' & __AUERLO_GetTxtExitcode(@exitCode) & '".', _
		@exitCode ? $AUERLO_LVL_FATAL : (@exitMethod >= $EXITCLOSE_BYLOGOFF ? $AUERLO_LVL_WARNING : $AUERLO_LVL_INFO), _
		True, _
		@exitCode, _
		@exitMethod, _
		@Compiled ? -1 : 0, _
		'' _
	)
	Return SetError($AUERLO_ERR_OK, 0, $AUERLO_RET_OK)
EndFunc

Func __AUERLO_HandleCOM(Const ByRef $oError, Const $sFilePath = @ScriptFullPath)
	Local Const $aItem = [ _
							'windescription: ' & $oError.windescription, _
							'description: '    & $oError.description, _
							'source: '         & $oError.source, _
							'helpfile: '       & $oError.helpfile, _
							'helpcontext: '    & $oError.helpcontext, _
							'lastdllerror: '   & $oError.lastdllerror _
						 ]

	_AUERLO_Log( _
		'__AUERLO_HandleCOM() - ' & _ArrayToString($aItem, '|'), _
		$AUERLO_LVL_ERROR, _
		True, _
		$oError.number, _
		$oError.retcode, _
		$oError.scriptline, _
		$sFilePath _
	)

	Return SetError($AUERLO_ERR_OK, 0, $AUERLO_RET_OK)
EndFunc

#EndRegion

#Region Output

Func __AUERLO_Output_File(Const $aData)
	Local Static $sCache = ''
	Local        $iErr   = 0

	$sCache &= StringFormat(__AUERLO_GetFormatLine(), _; _AUERLO_Read()'s $bHeader parameter and _AUERLO_View() assume $AUERLO_ATR__ENUM's declaration order.
							$aData[$AUERLO_ATR_DATETIME], _
							$aData[$AUERLO_ATR_LEVEL], _
							$aData[$AUERLO_ATR_FILENAME], _
							$aData[$AUERLO_ATR_LINE], _
							$aData[$AUERLO_ATR_ERROR], _
							$aData[$AUERLO_ATR_EXTENDED], _
							$aData[$AUERLO_ATR_MESSAGE] _
						   )

	Local $hFile = FileOpen($__g_aAUERLO_Opt[$AUERLO_SET_CUR][$AUERLO_OPT_FILEPATH], BitOR($FO_APPEND, $FO_CREATEPATH, $__g_aAUERLO_Opt[$AUERLO_SET_CUR][$AUERLO_OPT_FILEENC]))
	If $hFile = -1 Then Return SetError($AUERLO_ERR_FILEOPEN, 0, $AUERLO_RET_ERR)

	FileWriteLine($hFile, $sCache)
	$iErr = @error

	FileClose($hFile)

	If $iErr Then Return SetError($AUERLO_ERR_FILEWRITE, $iErr, $AUERLO_RET_ERR)

	$sCache = ''

	Return SetError($AUERLO_ERR_OK, 0, $AUERLO_RET_OK)
EndFunc

Func __AUERLO_Output_Std(Const $aData)
	Local       $sLine = ''
	Local Const $iAttr = @Compiled ? $AUERLO_ATR__ENUM - 2 : $AUERLO_ATR__ENUM
	Local       $aLine[$iAttr]

	If @Compiled Then

		$aLine[0] =                      $aData[$AUERLO_ATR_DATETIME]
		$aLine[1] = __AUERLO_GetTxtLevel($aData[$AUERLO_ATR_LEVEL])
		$aLine[2] =                  Hex($aData[$AUERLO_ATR_ERROR])
		$aLine[3] =                  Hex($aData[$AUERLO_ATR_EXTENDED])
		$aLine[4] =                      $aData[$AUERLO_ATR_MESSAGE]
		$sLine   &= StringFormat(__AUERLO_GetFormatLine($iAttr), $aLine[0], $aLine[1], $aLine[2], $aLine[3], $aLine[4])

	Else

		If $__g_aAUERLO_Opt[$AUERLO_SET_CUR][$AUERLO_OPT_STDCOLOR] Then $sLine = $__g_aAUERLO_ConPfx[ $aData[$AUERLO_ATR_LEVEL] ]

		$aLine[0] =                '"' & $aData[$AUERLO_ATR_FILENAME] & '"'
		$aLine[1] =                      $aData[$AUERLO_ATR_LINE]
		$aLine[2] = __AUERLO_GetTxtLevel($aData[$AUERLO_ATR_LEVEL])
		$aLine[3] =                  Hex($aData[$AUERLO_ATR_ERROR])
		$aLine[4] =                  Hex($aData[$AUERLO_ATR_EXTENDED])
		$aLine[5] =                      $aData[$AUERLO_ATR_DATETIME]
		$aLine[6] =                      $aData[$AUERLO_ATR_MESSAGE]
		$sLine   &= StringFormat(__AUERLO_GetFormatLine($iAttr), $aLine[0], $aLine[1], $aLine[2], $aLine[3], $aLine[4], $aLine[5], $aLine[6])

	EndIf

	If $aData[$AUERLO_ATR_LEVEL] >= $AUERLO_LVL_ERROR Then

		ConsoleWriteError($sLine)

	Else

		ConsoleWrite($sLine)

	EndIf

	Return SetError($AUERLO_ERR_OK, 0, $AUERLO_RET_OK)
EndFunc

Volatile Func __AUERLO_Output_Notify(Const $aData)

	Abs($aData[$AUERLO_ATR_ERROR]); AU3Check by-pass ($aData declared but not used).
	Beep(500, 100)
;	SoundPlay(@WindowsDir & '\media\notify.wav', $SOUND_NOWAIT)

	Return SetError($AUERLO_ERR_OK, 0, $AUERLO_RET_OK)
EndFunc

Func __AUERLO_Output_Prompt(Const $aData)
	Local $aResult[1][$AUERLO_ATR__ENUM]

	For $i1 = 0 To $AUERLO_ATR__ENUM - 1

		$aResult[0][$i1] = $aData[$i1]

	Next

	_AUERLO_View($aResult)
	If @error Then Return SetError(@error, @extended, $AUERLO_RET_ERR)

	Return SetError($AUERLO_ERR_OK, 0, $AUERLO_RET_OK)
EndFunc

Func __AUERLO_Output_WinEventLog(Const $aData)
	Local $iType = 0, _
		  $iCat  = 0, _
		  $iId   = 0
	Local $sUser = '';'Administrator';@UserName
	Local $aCont = [1, 0]
	Local $hLog  = 0

	Switch $aData[$AUERLO_ATR_LEVEL]

		Case $AUERLO_LVL_TRACE
			$iType = $AUERLO_WEL_TYPE_SUCCESS

		Case $AUERLO_LVL_DEBUG
			$iType = $AUERLO_WEL_TYPE_INFO

		Case $AUERLO_LVL_INFO
			$iType = $AUERLO_WEL_TYPE_INFO

		Case $AUERLO_LVL_WARNING
			$iType = $AUERLO_WEL_TYPE_WARNING

		Case $AUERLO_LVL_ERROR
			$iType = $AUERLO_WEL_TYPE_ERROR

		Case $AUERLO_LVL_FATAL
			$iType = $AUERLO_WEL_TYPE_ERROR

		Case Else
			$iType = $AUERLO_WEL_TYPE_ERROR

	EndSwitch

	$hLog = _EventLog__Open('', 'Application')
	If Not $hLog Then Return SetError($AUERLO_ERR_WELOPEN, 0, $AUERLO_RET_ERR)

	If Not _EventLog__Report($hLog, $iType, $iCat, $iId, $sUser, $aData[$AUERLO_ATR_MESSAGE], $aCont) Then _
	Return SetError($AUERLO_ERR_WELWRITE, 0, $AUERLO_RET_ERR)

	_EventLog__Close($hLog)

	Return SetError($AUERLO_ERR_OK, 0, $AUERLO_RET_OK)
EndFunc

#EndRegion
